import torch
from torch.optim.optimizer import Optimizer
from agc_optims.utils import agc

class AGC(Optimizer):
    r"""Implements adaptive gradient clipping independent from the optimizer.
    This implementation of AGC adapted from https://github.com/vballoli/nfnets-pytorch.
    Args:
        optimizer (Optimizer): Optimizer that inherits the Optimizer class from PyTorch
        clipping (float, optional): clipping value for the AGC (default: 1e-2)
        agc_eps (float, optional): term used in agc to prevent grads clipped to zero (default: 1e-3)
    .. _High-Performance Large-Scale Image Recognition Without Normalization 
        https://arxiv.org/abs/2102.06171
    """
    def __init__(self, optimizer, clipping=1e-2, agc_eps=1e-3):
        if not isinstance(optimizer, Optimizer):
            raise ValueError("No optimizer is given, inserted: {}".format(type(optimizer)))
        if not 0.0 <= clipping < 1.0:
            raise ValueError("Invalid clipping parameter: {}".format(clipping))
        if not 0.0 <= agc_eps:
            raise ValueError("Invalid agc_eps value: {}".format(agc_eps))
        self.optimizer = optimizer
        defaults = dict(clipping=clipping, agc_eps=agc_eps)
        defaults = {**self.optimizer.defaults, **defaults}
        param_groups = self.optimizer.__getstate__()['param_groups']
        super(AGC, self).__init__(param_groups, defaults)
    
    @torch.no_grad()
    def step(self, closure=None):
        
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            clipping = group['clipping']
            agc_eps = group['agc_eps']
            for p in group['params']:
                agc(param=p, clipping=clipping, eps=agc_eps)
        
        self.optimizer.step(closure)

        return loss



import torch

def unitwise_norm(x):
    if x.ndim <= 1:
        dim = 0
        keepdim = False
    elif x.ndim in [2, 3]:
        dim = 0
        keepdim = True
    elif x.ndim >= 4:
        dim = [i for i in range(1, x.ndim)]
        keepdim = True
    else:
        raise ValueError('Wrong input dimensions')
    return torch.sum(x**2, dim=dim, keepdim=keepdim) ** 0.5



def agc(param, clipping=1e-2, eps=1e-3):
    """
        Implements the adaptive gradient clipping according to the NFNet paper: https://arxiv.org/pdf/2102.06171.pdf
    """
    param_data = param.detach()
    grad_data = param.grad.detach()
    max_norm = unitwise_norm(param_data).clamp_(min=eps).mul_(clipping) # Clamp works faster than .max()! But iff min= is a float not a Tensor!
    grad_norm = unitwise_norm(grad_data)
    clipped_grad = grad_data.mul_((max_norm.div_(grad_norm.clamp(min=1e-6))))
    new_grads = torch.where(grad_norm < max_norm, grad_data, clipped_grad)
    param.grad.detach().copy_(new_grads)




import sys

import setuptools

sys.path.insert(0, "src")
import agc_optims

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="agc_optims",
    version=agc_optims.__version__,
    author="Fabian Deuser",
    description="Easy to use optimizers with adaptive gradient clipping. Written in PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Skyy93/agc_optims",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.0",
    install_requires=[
        "torch >= 1.6.0",
    ],
)